//////////////////////////////////////////////////////////////////////////
//
// FILE:   Interface.C
//
// DATE:   January 3, 1996
//
// AUTHOR:
//
//    James Kuffner, Jr
//    Computer Science Department
//    Stanford University
//    U.S.A.
//    email: kuffner@stanford.edu
//    http://robotics.stanford.edu/~kuffner
//
// DISTRIBUTION NOTICE
//
// Copyright (c) 1996 James Kuffner, Jr. All rights reserved.
//
// Permission to use, copy, modify and distribute this software for any
// purpose is hereby granted without fee, provided that the above
// copyright notice and this permission notice appear in all copies of
// this software and that you do not sell the software.
//
// THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
//////////////////////////////////////////////////////////////////////////

#ifndef IGNORE_DEPEND
#include <stdio.h>
#include <Xm/RowColumn.h>
#include <Xm/CascadeBG.h>
#include <Xm/Form.h>
#include <Xm/FileSB.h>
#include <Xm/PushBG.h>
#include <Xm/SeparatoG.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>
#include <Inventor/Xt/SoXt.h>
#endif

#include "Interface.h"
#include "Interpolate.h"


// menus
enum AnimMenuEntries {
    I_FILE = 0,
    I_FILE_QUIT,

    I_VIEW,
    I_VIEW_1,
    I_VIEW_2,
    I_VIEW_4,
    
    I_KEYFRAME,
    I_KEYFRAME_1,
    I_KEYFRAME_2,

    I_ANIMATE,
    I_ANIMATE_START,
    I_ANIMATE_STOP,

    I_ABOUT,
    I_ABOUT_ABOUT,

    I_MENU_NUM		// this must be the last entry
};

////////////////
//
// Define the structure of the menus.

static MenuItemInfo fileDataAM[] = {
  {"Quit",       I_FILE_QUIT,    PushButton, 0, "Aq"},
};

static MenuItemInfo viewDataAM[] = {
  {"One View",        I_VIEW_1,  PushButton, 0, "A1"},
  {"Two Views",       I_VIEW_2,  PushButton, 0, "A2"},
  {"Four Views",      I_VIEW_4,  PushButton, 0, "A4"},
};

static MenuItemInfo keyDataAM[] = {
  {"Set First Keyframe",      I_KEYFRAME_1,  PushButton, 0, "1"},
  {"Set Second Keyframe",     I_KEYFRAME_2,  PushButton, 0, "2"},
};

static MenuItemInfo animDataAM[] = {
  {"Start Playback",      I_ANIMATE_START,  PushButton, 0, "p"},
  {"Stop Playback",       I_ANIMATE_STOP,   PushButton, 0, "s"},
};

static MenuItemInfo helpDataAM[] = {
  {"About Interpolate...", I_ABOUT_ABOUT, PushButton, 0 ,"Ah" },
};


#define ITEMS(x) XtNumber(x), x

static MenuInfo menus[] = {
// {name, type, numItems, items}
   {"File",        PulldownMenu, ITEMS(fileDataAM) },
   {"View",        PulldownMenu, ITEMS(viewDataAM) },
   {"Keyframe",    PulldownMenu, ITEMS(keyDataAM) },
   {"Animate",     PulldownMenu, ITEMS(animDataAM) },
   {"About",       PulldownMenu, ITEMS(helpDataAM) },
};


//these are necessary for proper menu graying
enum AnimMenus {
  AMB_FILE = 0,
  AMB_VIEW,
  AMB_KEYFRAME,
  AMB_ANIMATE,
  AMB_HELP
};

#define HelpMenuIndex  AMB_HELP


//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////
Interface::Interface(Interpolate *interpolate, Widget parent)
{
  _interpolate = interpolate;
  _parentWindow = parent;

  // create a form widget to hold the interface elements
  _formWidget = XmCreateForm(_parentWindow, "form", NULL, 0);

  // create the menu bar
  _BuildMenus(XtNumber(menus), menus);
  menuBar->manage();
  XtManageChild(_formWidget);
}


//////////////////////////////////////////////////////////////////////////
// Destructor
//////////////////////////////////////////////////////////////////////////
Interface::~Interface()
{
  delete menuBar;
}


//////////////////////////////////////////////////////////////////////////
// ErrorDialog :  Brings up an error dialog
//////////////////////////////////////////////////////////////////////////
void Interface::ErrorDialog(const char *msg)
{
  char command[1000];
  sprintf(command, "xconfirm -c -t -header 'Error' '%s' -B 'OK' > /dev/null",
	  msg);
  system(command);
}


/////////////////////
// private methods
/////////////////////

//////////////////////////////////////////////////////////////////////////
// _BuildMenus
//////////////////////////////////////////////////////////////////////////
void Interface::_BuildMenus(int n, MenuInfo *menuInfoList)
{
  int i;
  MenuBarMenu *menuBarMenu;

  menuBar = new MenuBar(_formWidget);
  menuBar->indexWidgets(I_MENU_NUM);
  
  for (i = 0; i < n; i++) {
    menuBarMenu = menuBar->addMenu(menuInfoList[i]);
    menuBarMenu->menu->setCallback(_MenuHandler, this);
    if (i == HelpMenuIndex)
      XtVaSetValues(menuBar->menuBarWidget(), XmNmenuHelpWidget,
		    menuBarMenu->button, NULL);
  }
}


//////////////////////////////////////////////////////////////////////////
// _MenuHandler
//////////////////////////////////////////////////////////////////////////
void Interface::_MenuHandler(MenuItem *item, void *userData)
{
  ((Interface *) userData)->_HandleMenuEvent(item);
}


//////////////////////////////////////////////////////////////////////////
// _ShowAboutDialog :  Brings up the "About..." dialog
//////////////////////////////////////////////////////////////////////////
void Interface::_ShowAboutDialog()
{
  const char *s1 = "Rotation Keyframe Interpolation Demonstration";
  const char *s2 = "                      James Kuffner, Jr.";
  const char *s3 = "                        (January 1996)";
  const char *s4 = " ";
  const char *s5 = "               email: kuffner@stanford.edu";
  const char *s6 = "          http://robotics.stanford.edu/~kuffner";

  char command[1000];
  sprintf(command, "xconfirm -c -t '%s' -t '%s' -t '%s' -t '%s' -t '%s' -t '%s' -B 'OK' -header 'About Interpolate' > /dev/null", s1, s2, s3, s4, s5, s6);
  system(command);
}


//////////////////////////////////////////////////////////////////////////
// _FileQuit
//////////////////////////////////////////////////////////////////////////
void Interface::_FileQuit()
{
  exit(0);
}


///////////////////////////////////////////////////////////////////////
// _UpdateViewRadioButtons
///////////////////////////////////////////////////////////////////////
void Interface::_UpdateViewRadioButtons(int which)
{
  // Set the appropriate radio button for the view
  XmToggleButtonSetState(menuBar->widgetByIndex(which),TRUE,FALSE);
}

//////////////////////////////////////////////////////////////////////////
// _Handlemenuevent : Handle a menu event.
//////////////////////////////////////////////////////////////////////////
void Interface::_HandleMenuEvent(MenuItem *item)
{

  switch (item->id) {
    
    ///////////////////////////////// File 
    
  case I_FILE_QUIT:
    _FileQuit();
    break;	

    ///////////////////////////////// View

  case I_VIEW_1:
    _interpolate->SetNumViews(1);
    break;

  case I_VIEW_2:
    _interpolate->SetNumViews(2);
    break;

  case I_VIEW_4:
    _interpolate->SetNumViews(4);
    break;

    ///////////////////////////////// Keyframe

  case I_KEYFRAME_1:
    _interpolate->SetKeyframe(1);
    break;

  case I_KEYFRAME_2:
    _interpolate->SetKeyframe(2);
    break;

    ///////////////////////////////// Animate

  case I_ANIMATE_START:
    _interpolate->Playback(true);
    break;

  case I_ANIMATE_STOP:
    _interpolate->Playback(false);
    break;

    ///////////////////////////////// About

  case I_ABOUT_ABOUT:
    _ShowAboutDialog();
    break;
    
  default:
    ErrorDialog("Unknown menu option!");
    break;
  } 

}













