//////////////////////////////////////////////////////////////////////////
//
// FILE:   Interpolate.C
//
// DATE:   January 3, 1996
//
// AUTHOR:
//
//    James Kuffner, Jr
//    Computer Science Department
//    Stanford University
//    U.S.A.
//    email: kuffner@stanford.edu
//    http://robotics.stanford.edu/~kuffner
//
// DISTRIBUTION NOTICE
//
// Copyright (c) 1996 James Kuffner, Jr. All rights reserved.
//
// Permission to use, copy, modify and distribute this software for any
// purpose is hereby granted without fee, provided that the above
// copyright notice and this permission notice appear in all copies of
// this software and that you do not sell the software.
//
// THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
//////////////////////////////////////////////////////////////////////////

#ifndef IGNORE_DEPEND
#include <stdio.h>
#include <Xm/PanedW.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/SbLinear.h>
#endif

#include "Interpolate.h"
#include "Interface.h"
#include "Renderer.h"
#include "TextArea.h"

// weight increment
const float WEIGHT_INC      = 0.01;

// tolerancing
const float ZERO_THRESHHOLD = 0.0001;


//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////
Interpolate::Interpolate(Widget mainWindow)
{
  _mainWindow = mainWindow;
  _stopped = TRUE;
  _weight = 0.0;
  _k1set = false;
  _k2set = false;

  // ************** THE KEY-FRAME DATA ***************
  for (int i = 0; i < 4; i++) {
    _key1[i] = 0.0;
    _key2[i] = 0.0;
  }

  // ************** THE INTERFACE OBJECT ***************
  // This object handles menus and general GUI interface issues
  _interface = new Interface(this, _mainWindow);

  _pane = XtVaCreateWidget("pane",xmPanedWindowWidgetClass,
			   _interface->FormWidget(),
			   XmNtopAttachment, XmATTACH_WIDGET,
			   XmNtopWidget, _interface->MenuWidget(),
			   XmNtopOffset, 6,
			   XmNleftAttachment, XmATTACH_FORM,
			   XmNrightAttachment, XmATTACH_FORM,
			   XmNbottomAttachment, XmATTACH_FORM,
			   NULL);

  // **************** THE RENDER OBJECT **************
  // This object will create the viewer
  _mfr = new Renderer(_pane, this);

  // ************* THE TEXTAREA WIDGET ***************
  _textArea = new TextArea(_pane);
  _textArea->setCallback(TaWeightChange,(TaCallback)_WeightChangeCB, this);
  _textArea->setCallback(TaQuatChange,(TaCallback)_QuatChangeCB, this);
  _textArea->setCallback(TaAxisChange,(TaCallback)_AxisChangeCB, this);
  _textArea->setCallback(TaEulerChange,(TaCallback)_EulerChangeCB, this);
  _textArea->setCallback(TaUpdateValues,(TaCallback)_UpdateValuesCB, this);

  XtManageChild(_pane);
  XtManageChild(_interface->MenuWidget());
  XtManageChild(_interface->FormWidget());

  _textArea->InitMode();
  _textArea->SetWeight(_weight);
}



//////////////////////////////////////////////////////////////////////////
// Destructor : we most likely won't call this unless the program quits
//////////////////////////////////////////////////////////////////////////
Interpolate::~Interpolate()
{
  delete _mfr;
  delete _interface;
}


//////////////////////////////////////////////////////////////////////////
// ShowErrorDialog : Pop up a simple errror dialog, wait for response,
//                   and return.
//////////////////////////////////////////////////////////////////////////
void Interpolate::ShowErrorDialog(char *s)
{
  _interface->ErrorDialog(s);
}


//////////////////////////////////////////////////////////////////////////
// AdvanceFrame
//////////////////////////////////////////////////////////////////////////
void Interpolate::AdvanceFrame()
{
  //increment frame index
  _weight += WEIGHT_INC;

  if (_weight > 1.0)
    _weight = 0.0;
  
  _textArea->SetWeight(_weight);
  DrawFrame();
}

//////////////////////////////////////////////////////////////////////////
// DrawFrame
//////////////////////////////////////////////////////////////////////////
void Interpolate::DrawFrame()
{
  float q[4];
  float axisLength;
  float oneMinusWeight = 1.0 - _weight;
  int i;
  SbRotation rot;
  
  switch (_textArea->Mode()) {
  case QUAT:
    rot = SbRotation::slerp(SbRotation(_key1),SbRotation(_key2), _weight);
    rot.getValue(q[0],q[1],q[2],q[3]);
    _mfr->SetQuatTransform(q);
    _textArea->SetSliderValues(q);
    break;

  case AXIS:
    // linearly interpolate the axis components and angle separately
    for (i = 0; i < 4; i++)
      q[i] = oneMinusWeight*_key1[i] + _weight*_key2[i];
    // normalize the axis
    axisLength = sqrt(q[1]*q[1] + q[2]*q[2] + q[3]*q[3]);
    // NOTE: this code doesn't work for the special case of the axis
    // being interpolated through the origin.  In this case, we can
    // use either keyframe axis, but must remember to flip the sign of
    // the angle after the zero crossing.
    if (axisLength < ZERO_THRESHHOLD)
      for (i = 1; i < 4; i++)
	q[i] = _key1[i];
    else
      for (i = 1; i < 4; i++)
	q[i] *= 1.0/axisLength;

    // fix the angle if necessary
    while (q[0] > 180.0)
      q[0] -= 360.0;
    while (q[0] < -180.0)
      q[0] += 360.0;

    _mfr->SetAxisTransform(q);
    _textArea->SetSliderValues(q);
    break;

  case EULER:
    // linearly interpolate each angle
    for (i = 0; i < 3; i++) {
      q[i] = oneMinusWeight*_key1[i] + _weight*_key2[i];
      while (q[i] > 180.0)
	q[i] -= 360.0;
      while (q[i] < -180.0)
	q[i] += 360.0;
    }
    _mfr->SetEulerTransform(q);
    _textArea->SetSliderValues(q);
    break;

  default:
    ShowErrorDialog("Unknown mode");
    break;
  }
}


//////////////////////////////////////////////////////////////////////////
// SetNumViews
//////////////////////////////////////////////////////////////////////////
void Interpolate::SetNumViews(int n)
{
  _mfr->SetNumViews(n,false);
}


//////////////////////////////////////////////////////////////////////////
// SetKeyframe
//////////////////////////////////////////////////////////////////////////
void Interpolate::SetKeyframe(int n)
{
  float *q;
  int lim = 4;

  if (n == 1) {
    printf("Keyframe 1 set\n");
    _k1set = true;
    q = _key1;
  }
  else {
    printf("Keyframe 2 set\n");
    _k2set = true;
    q = _key2;
  }

  switch (_textArea->Mode()) {
  case QUAT:
    _mfr->GetQuatTransform(q);
    break;

  case AXIS:
    _mfr->GetAxisTransform(q);
    break;

  case EULER:
    _mfr->GetEulerTransform(q);
    _textArea->GetSliderValues(q);
    lim = 3;
    break;
  }

  for (int i = 0; i < lim; i++)
    printf("\tq[%d] = %f\n",i,q[i]);
  printf("\n");
}


//////////////////////////////////////////////////////////////////////////
// Playback
//////////////////////////////////////////////////////////////////////////
void Interpolate::Playback(bool start)
{
  if (_k1set && _k2set) {
    if (start) {
      if (_stopped) {
	_mfr->StartPlayback();
	_stopped = false;
      }
    }
    else {
      if (!_stopped) {
	_mfr->StopPlayback();
	_stopped = true;
      }
    }
  }
  else
    ShowErrorDialog("You must set two keyframes\nbefore you can interpolate.");
}


//////////////////////////////////////////////////////////////////////////
// UpdateValues : get rotation values from Inventor node and
//  update the slider display
//////////////////////////////////////////////////////////////////////////
void Interpolate::UpdateValues(RotMode mode)
{
  float q[4];

  // should convert keyframe values
  _k1set = false;
  _k2set = false;

  // stop playback if necessary
  if (!_stopped) {
    _mfr->StopPlayback();
    _stopped = true;
  }

  switch (mode) {
  case QUAT:
    _mfr->GetQuatTransform(q);
    break;

  case AXIS:
    _mfr->GetAxisTransform(q);
    break;

  case EULER:
    _mfr->GetEulerTransform(q);
    break;
  }

  _textArea->SetSliderValues(q);
}

//////////////////////////////////////////////////////////////////////////
// _WeightChangeCB
//////////////////////////////////////////////////////////////////////////
void Interpolate::_WeightChangeCB(TextArea *, float *val, TaPointer clientData)
{
  Interpolate* myself = (Interpolate *)clientData;
  myself->_weight = val[P_WEIGHT];
  if (myself->_k1set && myself->_k2set)
    myself->DrawFrame();
  else
    myself->ShowErrorDialog("You need to set keyframes\nbefore you can interpolate.");
}


//////////////////////////////////////////////////////////////////////////
// _QuatChangeCB
//////////////////////////////////////////////////////////////////////////
void Interpolate::_QuatChangeCB(TextArea *, float *val, TaPointer clientData)
{
  Interpolate* myself = (Interpolate *)clientData;
  int i;

  // normalize quaternion
  float imag = 1.0/(sqrt(val[0]*val[0] + val[1]*val[1] + val[2]*val[2]
			 + val[3]*val[3]));
  for (i = 0; i < 4; i++)
    val[i] *= imag;

  myself->_mfr->SetQuatTransform(val);
  myself->_textArea->SetSliderValues(val);
}

//////////////////////////////////////////////////////////////////////////
// _AxisChangeCB
//////////////////////////////////////////////////////////////////////////
void Interpolate::_AxisChangeCB(TextArea *, float *val, TaPointer clientData)
{
  Interpolate* myself = (Interpolate *)clientData;
  int i;

  // normalize axis
  float imag = 1.0/(sqrt(val[1]*val[1] + val[2]*val[2] + val[3]*val[3]));
  for (i = 1; i < 4; i++)
    val[i] *= imag;
  
  myself->_mfr->SetAxisTransform(val);
  myself->_textArea->SetSliderValues(val);
}

//////////////////////////////////////////////////////////////////////////
// _EulerChangeCB
//////////////////////////////////////////////////////////////////////////
void Interpolate::_EulerChangeCB(TextArea *, float *val, TaPointer clientData)
{
  Interpolate* myself = (Interpolate *)clientData;

  for (int i = 0; i < NUM_FIELDS; i++)
  myself->_mfr->SetEulerTransform(val);
}

//////////////////////////////////////////////////////////////////////////
// _UpdateValuesCB
//////////////////////////////////////////////////////////////////////////
void Interpolate::_UpdateValuesCB(TextArea *ta, const float *,
				  TaPointer clientData)
{
  Interpolate* myself = (Interpolate *)clientData;
  myself->UpdateValues(ta->Mode());
}





