//////////////////////////////////////////////////////////////////////////
//
// FILE:   Menu.C
//
// DATE:   January 3, 1996
//
// AUTHOR:
//
//    James Kuffner, Jr
//    Computer Science Department
//    Stanford University
//    U.S.A.
//    email: kuffner@stanford.edu
//    http://robotics.stanford.edu/~kuffner
//
// DISTRIBUTION NOTICE
//
// Copyright (c) 1996 James Kuffner, Jr. All rights reserved.
//
// Permission to use, copy, modify and distribute this software for any
// purpose is hereby granted without fee, provided that the above
// copyright notice and this permission notice appear in all copies of
// this software and that you do not sell the software.
//
// THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
//////////////////////////////////////////////////////////////////////////

#include "Menu.h"

#ifndef IGNORE_DEPEND
#include <string.h>
#include <stdio.h>
#endif

#include <Xm/RowColumn.h>
#include <Xm/CascadeBG.h>
#include <Xm/Form.h>
#include <Xm/FileSB.h>
#include <Xm/PushBG.h>
#include <Xm/SeparatoG.h>
#include <Xm/ToggleBG.h>

////////////////////////////////////////////////////////////////
//
// Methods for class Menu
//
////////////////////////////////////////////////////////////////

////////////////
//
// Construct a new empty menu of the given type.

Menu::Menu(Widget parent, char *name, MenuType type, int menuIndex)
{
  _menuIndex = menuIndex;
  _parent = parent;
  _type = type;
  _name = strdup(name);
  _items = NULL;

  _userData = NULL;
  _menuCallback = NULL;

  initialize();
}

////////////////
//
// Destroy a menu and its items.

Menu::~Menu()
{
  delete _name;
  XtDestroyWidget(_menu);
}

////////////////
//
// Set the user callback function for this menu.

void Menu::setCallback(MenuCallback callback, void *userData)
{
  _menuCallback = callback;
  _userData = userData;
}

////////////////
//
// Initialize an emtpy menu.

void Menu::initialize()
{
  switch (_type) {

  case PopupMenu:
    _menu = XmCreatePopupMenu(_parent, _name, NULL, 0);
    _option = NULL;
    break;

  case PulldownMenu:
    _menu = XmCreatePulldownMenu(_parent, _name, NULL, 0);
    _option = NULL;
    break;

  case OptionMenu:
    XmString s = XmStringCreateLocalized(_name);

    _menu = XmCreatePulldownMenu(_parent, _name, NULL, 0);
    _option = XmCreateOptionMenu(_parent, _name, NULL, 0);
    XtVaSetValues(_option, XmNsubMenuId, _menu, XmNlabelString, s, NULL);

    XmStringFree(s);
    break;
  }
}

////////////////
//
// Add an item to the menu.

MenuItem *Menu::addItem(const MenuItemInfo &itemInfo)
{
  MenuItem *item;
  String callbackType = NULL, name;
  WidgetClass widgetClass;

  // set up a new menu item structure

  item = new MenuItem;
  item->menu = this;
  item->id = itemInfo.id;

  name = itemInfo.name;

  // figure out the parameters of the widget to create

  switch (itemInfo.type) {

  case Separator:
    name = NULL;
    widgetClass = xmSeparatorGadgetClass;
    break;

  case PushButton:
    widgetClass = xmPushButtonGadgetClass;
    callbackType = XmNactivateCallback;
    break;

  case ToggleButton:
  case RadioButton:
    widgetClass = xmToggleButtonGadgetClass;
    callbackType = XmNvalueChangedCallback;
    break;

  case SubMenu:
    break;
  }

  // create the widget for the menu item

  item->widget = XtCreateWidget(name, widgetClass, _menu, NULL, 0);

  if (itemInfo.type == RadioButton)
    // if the button is a radio button, set the radio constraint
    XtVaSetValues(item->widget, XmNindicatorType, XmONE_OF_MANY, NULL);

  // check for a keyboard accelerator

  if (itemInfo.accelKey) {
    char show[64], accel[64];
    XmString accelText;
    int i, length;

    length = strlen(itemInfo.accelKey);
    strcpy(show, "");
    strcpy(accel, "");

    for (i = 0; i < length - 1; i++)
      switch (itemInfo.accelKey[i]) {

      case 'A': // Alt
	strcat(show, "Alt+");
	strcat(accel, "Alt ");
	break;

      case 'S': // Shift
	strcat(show, "Shift+");
	strcat(accel, "Shift ");
	break;

      case 'C': // Ctrl
	strcat(show, "Ctrl+");
	strcat(accel, "Ctrl ");

      }

    strcat(show, itemInfo.accelKey + length - 1);
    strcat(accel, "<Key> ");
    strcat(accel, itemInfo.accelKey + length - 1);

    accelText = XmStringCreateLocalized(show);

    XtVaSetValues(item->widget, XmNaccelerator, accel,
		  XmNacceleratorText, accelText, NULL);

    XmStringFree(accelText);
  }

  // if the button is a toggle or radio, set its initial state

  if ((itemInfo.type == ToggleButton) || (itemInfo.type == RadioButton))
    XmToggleButtonGadgetSetState(item->widget, itemInfo.initValue, False);

  // add a callback for the menu item and manage the widget

  if (callbackType)
    XtAddCallback(item->widget, callbackType, callback, (XtPointer) item);

  XtManageChild(item->widget);

  // add the new menu item to the internal list

  item->next = _items;
  _items = item;

  return item;
}

////////////////
//
// Add a set of items to the menu.

void Menu::addItems(int n, MenuItemInfo *itemInfoList)
{
  int i;

  for (i = 0; i < n; i++)
    addItem(itemInfoList[i]);
}

////////////////
//
// Remove the specified item from the menu.

void Menu::removeItem(MenuItem *item)
{
  MenuItem *ptr;

  XtDestroyWidget(item->widget);

  // remove the item from the internal list

  if (_items == item)
    _items = item->next;

  else
    for (ptr = _items; ptr->next; ptr = ptr->next)
      if (ptr->next == item) {
	ptr->next = item->next;
	break;
      }
}

////////////////
//
// Remove all of the items from the menu.

void Menu::removeAllItems()
{
  MenuItem *ptr, *next = 0;

  for (ptr = _items; ptr; ptr = next) {
    next = ptr->next;
    XtDestroyWidget(ptr->widget);
    delete ptr;
  }

  _items = NULL;
}

////////////////
//
// Set the currently selected item in the (option) menu, given
// the id of the item.

void Menu::selectItem(int id)
{
  MenuItem *ptr;

  if (_type != OptionMenu) return;
  
  for (ptr = _items; ptr; ptr = ptr->next)
    if (ptr->id == id) {
      XtVaSetValues(_option, XmNmenuHistory, ptr->widget, NULL);
      break;
    }
}

////////////////
//
// Get the id of the currently selected item in the (option) menu.
// Returns -1 if there is no selected item, which should never happen.

int Menu::selectedItem()
{
  MenuItem *ptr;
  Widget w;

  if (_type != OptionMenu) return -1;

  XtVaGetValues(_option, XmNmenuHistory, &w, NULL);

  for (ptr = _items; ptr; ptr = ptr->next)
    if (ptr->widget == w)
      return ptr->id;

  return -1;
}

////////////////
//
// Static member function to set the label of the given widget.

void Menu::setLabel(Widget w, const char *label)
{
  XmString s;

  s = XmStringCreateLocalized(String(label));
  XtVaSetValues(w, XmNlabelString, s, NULL);
  XmStringFree(s);
}

////////////////
//
// Static callback function called when a menu item is activated.

void Menu::callback(Widget, XtPointer client, XtPointer)
{
  MenuItem *item = (MenuItem *) client;

  if (item->menu->_menuCallback)
    item->menu->_menuCallback(item, item->menu->_userData);
}

////////////////////////////////////////////////////////////////
//
// Methods for class MenuBar
//
////////////////////////////////////////////////////////////////

////////////////
//
// Constructor

MenuBar::MenuBar(Widget parent)
{
  _parent = parent;

  _menuBar = XmCreateMenuBar(_parent, "menuBar", NULL, 0);
  XtVaSetValues(_menuBar,
		XmNtopAttachment, XmATTACH_FORM,
		XmNleftAttachment, XmATTACH_FORM,
		XmNrightAttachment, XmATTACH_FORM,
		NULL);

  _numMenus = 0;
  _menus = NULL;
  _indexedWidgets = NULL;
  _numItems = 0;
}

////////////////
//
// Destructor

MenuBar::~MenuBar()
{
  if (_indexedWidgets)
    delete [] _indexedWidgets;

  XtDestroyWidget(_menuBar);
}

////////////////
//
// Allocate space for an array of the menu items' widgets indexed
// by their ID.

void MenuBar::indexWidgets(int numItems)
{
  if (!_indexedWidgets) {
    _indexedWidgets = new Widget[numItems];
    _numItems = numItems;
  }
}

////////////////
//
// Return the widget for the menu item with the given ID.

Widget MenuBar::widgetByIndex(int id)
{
  if (id >= 0 && id < _numItems)
    return _indexedWidgets[id];

  return 0;
}

////////////////
//
// Set the sensitivity of the menu item specified by the given ID.

void MenuBar::enable(int id, int flag)
{
  XtSetSensitive(widgetByIndex(id), flag);
}

////////////////
//
// Add a new menu, specified by an MenuInfo structure, to the
// menu bar.

MenuBarMenu *MenuBar::addMenu(const MenuInfo &menuInfo)
{
  MenuBarMenu *menuBarMenu;
  Menu *menu;
  Widget button;

  // create a menu

  menu = new Menu(_menuBar, menuInfo.name, menuInfo.type, _numMenus);

  // create a cascade button for the menu

  button = XtCreateWidget(menuInfo.name, xmCascadeButtonGadgetClass,
			  _menuBar, NULL, 0);
  XtVaSetValues(button, XmNsubMenuId, menu->menuWidget(), NULL);
  XtManageChild(button);

  // add the items to the menu

  menu->addItems(menuInfo.numItems, menuInfo.items);

  // if we are storing an indexed widget array, add the new items

  if (_indexedWidgets)
    for (MenuItem *item = menu->_items; item; item = item->next)
      _indexedWidgets[item->id] = item->widget;

  // add a new node to the internal list of menus

  menuBarMenu = new MenuBarMenu;
  menuBarMenu->menu = menu;
  menuBarMenu->button = button;
  menuBarMenu->next = _menus;
  _menus = menuBarMenu;
  _numMenus++;

  return menuBarMenu;
}

////////////////
//
// Enable or disable a menu's cascade button.

void MenuBar::enableMenu(MenuBarMenu *menuBarMenu, int flag)
{
  XtSetSensitive(menuBarMenu->button, flag);
}

void MenuBar::enableMenu(int menuIndex, int flag)
{
  int i;
  MenuBarMenu *menu;

  menuIndex = _numMenus - menuIndex - 1;
  for (i = 0, menu = _menus; menu && i < menuIndex; i++, menu = menu->next);
  if (menu) enableMenu(menu, flag);
}

////////////////
//
// Set the label of a particular menu.

void MenuBar::setLabel(int id, const char *label)
{
  Menu::setLabel(widgetByIndex(id), label);
}

////////////////
//
// Set the state of a toggle button menu item.

void MenuBar::setState(int id, int flag)
{
  XmToggleButtonGadgetSetState(widgetByIndex(id), flag, False);
}

////////////////
//
// Manage the menu bar.

void MenuBar::manage()
{
  XtManageChild(_menuBar);
}
