//////////////////////////////////////////////////////////////////////////
//
// FILE:   Menu.h
//
// DATE:   January 3, 1996
//
// AUTHOR:
//
//    James Kuffner, Jr
//    Computer Science Department
//    Stanford University
//    U.S.A.
//    email: kuffner@stanford.edu
//    http://robotics.stanford.edu/~kuffner
//
// DISTRIBUTION NOTICE
//
// Copyright (c) 1996 James Kuffner, Jr. All rights reserved.
//
// Permission to use, copy, modify and distribute this software for any
// purpose is hereby granted without fee, provided that the above
// copyright notice and this permission notice appear in all copies of
// this software and that you do not sell the software.
//
// THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
//////////////////////////////////////////////////////////////////////////

#ifndef MENU_H
#define MENU_H

#include <Xm/Xm.h>

////////////////
//
// Valid menu types.

enum MenuType {
  PulldownMenu, PopupMenu, OptionMenu
};

////////////////
//
// Valid menu item types.

enum MenuItemType {
  Separator, PushButton, ToggleButton, RadioButton, SubMenu
};

////////////////////////////////////////////////////////////////
//
// Structures used when specifying menus.
//
////////////////////////////////////////////////////////////////

////////////////
//
// Structure containing information to construct a single menu item.

struct MenuItemInfo {
  char *name;
  int id;
  int type;
  int initValue;
  char *accelKey;
};

////////////////
//
// Structure containing information to construct a menu and all of
// its items.

struct MenuInfo {
  char *name;
  MenuType type;
  int numItems;
  struct MenuItemInfo *items;
};

////////////////////////////////////////////////////////////////
//
// Structures and classes used to represent menus.
//
////////////////////////////////////////////////////////////////

////////////////
//
// Structure to represent a single menu item.

struct MenuItem {
  int id;
  Widget widget;
  class Menu *menu;
  MenuItem *next;
};

typedef MenuItem *MenuItemPtr;

typedef void (*MenuCallback)(MenuItem *item, void *userData);

////////////////
//
// Class to represent a menu and all its items.

class Menu {
  friend class MenuBar;

public:
  Menu(Widget parent, char *name, MenuType type, int menuIndex = 0);
  ~Menu();

  void setCallback(MenuCallback callback, void *userData = NULL);

  MenuItem *addItem(const MenuItemInfo &itemInfo);
  void addItems(int n, MenuItemInfo *itemInfoList);

  Widget menuWidget() { return _menu; }
  Widget optionWidget() { return _option; }

  void removeItem(MenuItem *item);
  void removeAllItems();

  void selectItem(int id);
  int selectedItem();

  static void setLabel(Widget widget, const char *label);

private:
  static void callback(Widget, XtPointer, XtPointer);

  void initialize();

  char *_name;
  MenuType _type;
  Widget _menu, _parent, _option;
  int _menuIndex;
  MenuItem *_items;
  void *_userData;
  MenuCallback _menuCallback;
};

////////////////////////////////////////////////////////////////
//
// Structures and classes used to represent menu bars.
//
////////////////////////////////////////////////////////////////

////////////////
//
// Structure to represent a single menu on the menu bar.

struct MenuBarMenu {
  Menu *menu;
  Widget button;
  MenuBarMenu *next;
};

////////////////
//
// Class to represent a menu bar.

class MenuBar {
public:
  MenuBar(Widget parent);
  ~MenuBar();

  void indexWidgets(int numItems);
  Widget widgetByIndex(int id);

  void enable(int id, int flag);

  MenuBarMenu *addMenu(const MenuInfo &menuInfo);

  void enableMenu(MenuBarMenu *menuBarMenu, int flag);
  void enableMenu(int menuIndex, int flag);

  void setLabel(int id, const char *label);
  void setState(int id, int flag);

  void manage();

  Widget menuBarWidget() { return _menuBar; }

private:
  Widget _menuBar, _parent;
  Widget *_indexedWidgets;
  MenuBarMenu *_menus;
  int _numMenus, _numItems;
};

#endif // MENU_H
