//////////////////////////////////////////////////////////////////////////
//
// FILE:   Renderer.C
//
// DATE:   January 3, 1996
//
// AUTHOR:
//
//    James Kuffner, Jr
//    Computer Science Department
//    Stanford University
//    U.S.A.
//    email: kuffner@stanford.edu
//    http://robotics.stanford.edu/~kuffner
//
// DISTRIBUTION NOTICE
//
// Copyright (c) 1996 James Kuffner, Jr. All rights reserved.
//
// Permission to use, copy, modify and distribute this software for any
// purpose is hereby granted without fee, provided that the above
// copyright notice and this permission notice appear in all copies of
// this software and that you do not sell the software.
//
// THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
//////////////////////////////////////////////////////////////////////////

#ifndef IGNORE_DEPEND
#include <assert.h>
#include <Xm/DrawingA.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoSelection.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoLineSet.h>
#include <Inventor/sensors/SoTimerSensor.h>
#include <Inventor/fields/SoMFLong.h>
#include <Inventor/SoDB.h>
#endif

#include "Renderer.h"
#include "Interpolate.h"


//degree-radian conversion
#define  DEG2RAD     (M_PI/180.0)
#define  RAD2DEG     (180.0/M_PI)
#define  VERY_SMALL  0.0001

//the grid lines
#define  GRID_MIN   -50.0
#define  GRID_MAX    50.0
#define  GRID_LINES  100

//viewer defaults
#define  MIN_VIEWER_HEIGHT      100
#define  DEFAULT_VIEWER_WIDTH   800
#define  DEFAULT_VIEWER_HEIGHT  500
#define  DEFAULT_NUM_VIEWS      1

// the timer sensor interval
const float TIMER_SENSOR_INTERVAL = (1.0/20.0);  // 20 fps

//the viewers
enum { PERSPECTIVE=0, TOP, LEFT, FRONT};

//default positions of viewer cameras at startup
static float defaultCam[NUM_VIEWERS][3] = {{ 1.5, 1.0, 2.5 },
					   { 0.0, 4.0, 0.0 },
					   { 4.0, 1.0, 0.0 },
					   { 0.0, 1.0, 4.0 }};
static char *viewerName[NUM_VIEWERS] = { "Perspective View",
					 "Top View",
					 "Left View",
					 "Front View" };


// the model to load into the scene
const char *MODEL_FILENAME  = "X29.iv";
const char *AXES_FILENAME   = "axes.iv";


//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////
Renderer::Renderer(Widget parent, Interpolate *interpolate)
{
  _interpolate = interpolate;

  //create the root
  _ivRoot = new SoSelection;
  _ivRoot->ref();
  _ivRoot->policy.setValue(SoSelection::SINGLE);

  // *********** THE RENDER AREA ***************
  // This will create a drawing area widget for the inventor viewers
  _renderArea = XtVaCreateWidget("renderArea" ,xmDrawingAreaWidgetClass,
				 parent, NULL);
  XtVaSetValues(_renderArea, 
		XmNwidth, DEFAULT_VIEWER_WIDTH,
		XmNheight, DEFAULT_VIEWER_HEIGHT,
		XmNmarginWidth, 0,
		XmNmarginHeight, 0,
		XmNpaneMinimum, MIN_VIEWER_HEIGHT,
		NULL);
  XtAddCallback(_renderArea, XmNresizeCallback,
  		(XtCallbackProc) _ResizeRenderAreaCB, this);
  
  // Create the examiner viewers
  for (int v = 0; v < NUM_VIEWERS; v++) {
    _viewer[v] = new SoXtExaminerViewer(_renderArea, viewerName[v]);
    _viewer[v]->setSceneGraph(_ivRoot);
  }

  // Set default views for the cameras
  SoCamera *camera;
  for (v = 0; v < NUM_VIEWERS; v++) {
    camera = _viewer[v]->getCamera();
    camera->position.setValue(defaultCam[v][0],defaultCam[v][1],
			      defaultCam[v][2]);
    camera->pointAt(SbVec3f(0.0,0.0,0.0));
    _viewer[v]->saveHomePosition();
  }
  
  //create inventor scene graph nodes
  _ivGrid = new SoSeparator;
  _ivGridSwitch = new SoSwitch;

  //build Inventor scene graph hierarchy
  _ivRoot->addChild(_ivGridSwitch);
  _ivGridSwitch->addChild(_ivGrid);

  //display grid by default
  _displayGrid = true;
  _ivGridSwitch->whichChild.setValue(SO_SWITCH_ALL); 

  // build the render area
  _currentView = PERSPECTIVE;
  _numViews = DEFAULT_NUM_VIEWS;
  XtManageChild(_renderArea);
  SetNumViews(_numViews, FALSE);

  _InitGrid();
  _InitAxes();
  _InitPlayback();
  _InitModel();
}


//////////////////////////////////////////////////////////////////////////
// Destructor
//////////////////////////////////////////////////////////////////////////
Renderer::~Renderer()
{
  StopPlayback();
  delete _timerSensor;

  // Inventor should take care of destroying all other nodes (I think!)
  _ivRoot->unref();
}


//////////////////////////////////////////////////////////////////////////
// SetQuatTransform
//////////////////////////////////////////////////////////////////////////
void Renderer::SetQuatTransform(const float q[4])
{
  _ivTransform->rotation.setValue(q);
}

//////////////////////////////////////////////////////////////////////////
// GetQuatTransform
//////////////////////////////////////////////////////////////////////////
void Renderer::GetQuatTransform(float q[4])
{
  const int X=0, Y=1, Z=2, W=3;
  const SbRotation &rot = _ivTransform->rotation.getValue();
  rot.getValue(q[X],q[Y],q[Z],q[W]);
}

//////////////////////////////////////////////////////////////////////////
// SetAxisTransform
//////////////////////////////////////////////////////////////////////////
void Renderer::SetAxisTransform(const float q[4])
{
  SbVec3f axis(q[1],q[2],q[3]);
  float angle = DEG2RAD*q[0];
  _ivTransform->rotation.setValue(axis, angle);
}

//////////////////////////////////////////////////////////////////////////
// GetAxisTransform
//////////////////////////////////////////////////////////////////////////
void Renderer::GetAxisTransform(float q[4])
{
  SbVec3f axis;
  float angle;
  _ivTransform->rotation.getValue(axis, angle);
  axis.getValue(q[1],q[2],q[3]);
  q[0] = RAD2DEG*angle;
}


//////////////////////////////////////////////////////////////////////////
// SetEulerTransform
//////////////////////////////////////////////////////////////////////////
void Renderer::SetEulerTransform(const float euler[4])
{
  float quat[4],q[3],cy,cp,cr,sy,sp,sr;

  const int ROLL = 0, PITCH = 1, YAW = 2;
  const int X=0, Y=1, Z=2, W=3;

  // halve and convert to radians
  for (int i=0; i<3; i++)
    q[i] = 0.5 * DEG2RAD * euler[i];

  cr = cos(q[ROLL]);   //roll  = psi   = x-axis
  sr = sin(q[ROLL]);
  cp = cos(q[PITCH]);  //pitch = theta = y-axis
  sp = sin(q[PITCH]);
  cy = cos(q[YAW]);    //yaw   = phi   = z-axis
  sy = sin(q[YAW]);

  quat[W] = cr*cp*cy + sr*sp*sy;
  quat[X] = sr*cp*cy - cr*sp*sy;
  quat[Y] = cr*sp*cy + sr*cp*sy;
  quat[Z] = cr*cp*sy - sr*sp*cy;
  
  _ivTransform->rotation.setValue(quat);
}

//////////////////////////////////////////////////////////////////////////
// GetEulerTransform
//////////////////////////////////////////////////////////////////////////
void Renderer::GetEulerTransform(float euler[4])
{
  float q[4];
  const int X=0, Y=1, Z=2, W=3;
  const int ROLL = 0, PITCH = 1, YAW = 2;
  const SbRotation &rot = _ivTransform->rotation.getValue();
  rot.getValue(q[X],q[Y],q[Z],q[W]);

  // here, we first convert the quaternion to a column matrix,
  // and then extract the euler angles.  
  //
  // working with euler angles is inherently messy, since they
  // suffer from singularities and 'gimbal lock'.  Thus, converting
  // reliably is difficult at best.
  //
  // the original source for much of this code is from 'quatlib', the
  // quaternion library by Mark Livingston.

  double M[3][3];  // column matrix
  double xs,ys,zs,wx,wy,wz,xx,xy,xz,yy,yz,zz;

  xs = q[X] * 2.0; ys = q[Y] * 2.0; zs = q[Z] * 2.0;
  wx = q[W] * xs;  wy = q[W] * ys;  wz = q[W] * zs;
  xx = q[X] * xs;  xy = q[X] * ys;  xz = q[X] * zs;
  yy = q[Y] * ys;  yz = q[Y] * zs;  zz = q[Z] * zs;

  M[X][X] = 1.0 - (yy + zz);
  M[X][Y] = xy - wz;
  M[X][Z] = xz + wy;

  M[Y][X] = xy + wz;
  M[Y][Y] = 1.0 - (xx + zz);
  M[Y][Z] = yz - wx;

  M[Z][X] = xz - wy;
  M[Z][Y] = yz + wx;
  M[Z][Z] = 1.0 - (xx + yy);


  // now extract the euler angles
  double sinPitch, cosPitch, sinRoll, cosRoll, sinYaw, cosYaw;
  sinPitch = -M[2][0];
  cosPitch = sqrt(1 - sinPitch*sinPitch);
  
  if ( fabs(cosPitch) > VERY_SMALL ) {
    sinRoll = M[2][1] / cosPitch;
    cosRoll = M[2][2] / cosPitch;
    sinYaw  = M[1][0] / cosPitch;
    cosYaw  = M[0][0] / cosPitch;
  } 
  else {
    sinRoll = -M[1][2];
    cosRoll = M[1][1];
    sinYaw = 0;
    cosYaw = 1;
  }
  
  // get the angles
  euler[YAW]   = RAD2DEG * atan2(sinYaw, cosYaw);
  euler[PITCH] = RAD2DEG * atan2(sinPitch, cosPitch);
  euler[ROLL]  = RAD2DEG * atan2(sinRoll, cosRoll);
}


//////////////////////////////////////////////////////////////////////////
// SetNumViews
//////////////////////////////////////////////////////////////////////////
void Renderer::SetNumViews(int numViews, bool changeView)
{
  Dimension w,h;
  int v;
  XtVaGetValues(_renderArea, XmNwidth, &w, XmNheight, &h, NULL);

  //hide all viewers
  for (v = 0; v < 4; v++)
    _viewer[v]->hide();

  switch (numViews) {
  case 1:
    if ((changeView) && (numViews == _numViews)) {
      // cycle viewers
      _currentView = (_currentView + 1) % 4;
    }
    _numViews = 1;
    XtVaSetValues(_viewer[_currentView]->getWidget(),
		  XmNx, 0, XmNy, 0, NULL);
    _viewer[_currentView]->setSize(SbVec2s(w,h));
    _viewer[_currentView]->show();
    break;
    
  case 2:
    if ((changeView) && (numViews == _numViews)) {
      // cycle viewers
      if ((_currentView == PERSPECTIVE)||(_currentView == TOP))
	_currentView = LEFT;
      else
	_currentView = PERSPECTIVE;
    }
    else
      _currentView = (_currentView / 2)*2;
    _numViews = 2;
    XtVaSetValues(_viewer[_currentView]->getWidget(),
		  XmNx, w/2, XmNy, 0, NULL);
    XtVaSetValues(_viewer[_currentView + 1]->getWidget(),
		  XmNx, 0, XmNy, 0, NULL);
    _viewer[_currentView]->setSize(SbVec2s(w/2,h));
    _viewer[_currentView+1]->setSize(SbVec2s(w/2,h));
    _viewer[_currentView]->show();
    _viewer[_currentView+1]->show();
    break;
    
  case 4:
    _numViews = 4;
    XtVaSetValues(_viewer[TOP]->getWidget(),
		  XmNx, 0, XmNy, 0, NULL);
    XtVaSetValues(_viewer[PERSPECTIVE]->getWidget(),
		  XmNx, w/2, XmNy, 0, NULL);
    XtVaSetValues(_viewer[FRONT]->getWidget(),
		  XmNx, 0, XmNy, h/2, NULL);
    XtVaSetValues(_viewer[LEFT]->getWidget(),
		  XmNx, w/2, XmNy, h/2, NULL);
    for (v = 0; v < 4; v++) {
      _viewer[v]->setSize(SbVec2s(w/2,h/2));
      _viewer[v]->show();
    }
    break;
    
  default:
    fprintf(stderr, "ERROR: Unknown number of views %d\n", numViews);
    exit(-1);
    break;
  }
}

//////////////////////////////////////////////////////////////////////////
// StartPlayback : called when timeline button "play" is clicked.
//////////////////////////////////////////////////////////////////////////
void Renderer::StartPlayback()
{
  _timerSensor->schedule(); 
}


//////////////////////////////////////////////////////////////////////////
// StopPlayback : called when timeline button "stop" is clicked.
//////////////////////////////////////////////////////////////////////////
void Renderer::StopPlayback()
{
  if (_timerSensor->isScheduled())
    _timerSensor->unschedule(); 
}


//////////////////////////////////////////////////////////////////////////
// ToggleDisplay : change the state of a toggle switch
//////////////////////////////////////////////////////////////////////////
void Renderer::ToggleDisplay(int which)
{
  switch (which) {
  case TOGGLE_GRID:
    _displayGrid = !_displayGrid;
    _ivGridSwitch->whichChild.setValue(_displayGrid ? SO_SWITCH_ALL :
				       SO_SWITCH_NONE);
    break;

  default:
    fprintf(stderr, "BUG: Unknown toggle switch!\n");
    break;
  }
}


//////////////////////////////////////////////////////////////////////////
// AddSceneRoot
//////////////////////////////////////////////////////////////////////////
void Renderer::AddSceneRoot(SoSeparator *sceneRoot)
{
  _ivRoot->addChild(sceneRoot);
}

//////////////////////////////////////////////////////////////////////////
// RemoveSceneRoot
//////////////////////////////////////////////////////////////////////////
void Renderer::RemoveSceneRoot(SoSeparator *sceneRoot)
{
  _ivRoot->removeChild(sceneRoot);
}



/////////////////////
// private methods //
/////////////////////


//////////////////////////////////////////////////////////////////////////
// _PlayCB : callback for timer sensor
//////////////////////////////////////////////////////////////////////////
void Renderer::_PlayCB(void *userData, SoSensor *)
{
  Renderer *r = (Renderer *) userData;
  r->_interpolate->AdvanceFrame();
}


//////////////////////////////////////////////////////////////////////////
// _ResizeRenderAreaCB
//////////////////////////////////////////////////////////////////////////
void Renderer::_ResizeRenderAreaCB(Widget, void *userData, XtPointer)
{
  Renderer *myself = (Renderer *)(userData);
  myself->SetNumViews(myself->_numViews, FALSE);
}


/////////////////////////////////////////////////////////////////////////
// _InitGrid
//////////////////////////////////////////////////////////////////////////
void Renderer::_InitGrid()
{
  float xyz[4*(GRID_LINES+1)][3];
  int32_t  nPts[2*(GRID_LINES+1)];
  const int numLines = 2*(GRID_LINES+1);
  const float lineInterval = (GRID_MAX-GRID_MIN)/((float)GRID_LINES);

  //creates grid lines in the world
  SoMaterial *ivMaterial = new SoMaterial;
  SoCoordinate3 *ivCoord = new SoCoordinate3;
  SoLineSet *gridLineSet = new SoLineSet;

  ivMaterial->set("diffuseColor 0.1 0.2 1.0");
  _ivGrid->addChild(ivMaterial);
  _ivGrid->addChild(ivCoord);
  _ivGrid->addChild(gridLineSet);

  int i=0;
  for (int line=0; line<=GRID_LINES; line++) {
    //first endpoint
    xyz[i+2][2] = xyz[i][0] = GRID_MIN + lineInterval*line;
    xyz[i+2][1] = xyz[i][1] = 0.0;
    xyz[i+2][0] = xyz[i][2] = GRID_MIN;
    //second endpoint
    xyz[i+3][2] = xyz[i+1][0] = GRID_MIN + lineInterval*line;
    xyz[i+3][1] = xyz[i+1][1] = 0.0;
    xyz[i+3][0] = xyz[i+1][2] = GRID_MAX;
    i += 4;
  }
  for (i=0; i<numLines; i++) {
    //number of points connecting each line
    nPts[i] = 2;
  }
  SoMFVec3f *ivMFVec3f = (SoMFVec3f *) ivCoord->getField("point");
  ivMFVec3f->setValues(0,2*numLines,xyz);
  SoMFLong *ivMFLong = (SoMFLong *) gridLineSet->getField("numVertices");
  ivMFLong->setValues(0,numLines,nPts);
}


//////////////////////////////////////////////////////////////////////////
// _InitAxes
//////////////////////////////////////////////////////////////////////////
void Renderer::_InitAxes()
{
    // Add the model.
  SoSeparator *axes = _ReadFile(AXES_FILENAME);
  _ivRoot->addChild(axes);
}

//////////////////////////////////////////////////////////////////////////
// _InitPlayback
//////////////////////////////////////////////////////////////////////////
void Renderer::_InitPlayback()
{
  _timerSensor = new SoTimerSensor(_PlayCB, this);
  _timerSensor->unschedule();
  _timerSensor->setInterval(SbTime(TIMER_SENSOR_INTERVAL));
}


//////////////////////////////////////////////////////////////////////////
// _InitModel
//////////////////////////////////////////////////////////////////////////
void Renderer::_InitModel()
{
  SoSeparator *root = new SoSeparator();
  _ivRoot->addChild(root);

  // Add a Transform
  _ivTransform = new SoTransform;
  _ivTransform->scaleFactor.setValue(0.095,0.095,0.095);
  root->addChild(_ivTransform);
  
  // Add a constant model transform
  SoTransform *trans = new SoTransform;
  trans->rotation.setValue(0.5,0.5,0.5,0.5);

  // Add the model.
  SoSeparator *plane = _ReadFile(MODEL_FILENAME);
  root->addChild(trans);
  root->addChild(plane);
}


//////////////////////////////////////////////////////////////////////////
// _ReadFile
//////////////////////////////////////////////////////////////////////////
SoSeparator *Renderer::_ReadFile(const char *filename)
{
  // Open the input file
  SoInput mySceneInput;
  if (!mySceneInput.openFile(filename)) {
    fprintf(stderr, "Cannot open file %s\n", filename);
    return NULL;
  }
  // Read the whole file into the database
  SoSeparator *myGraph = SoDB::readAll(&mySceneInput);
  if (myGraph == NULL) {
    fprintf(stderr, "Problem reading file\n");
    return NULL;
  } 
  mySceneInput.closeFile();
  return myGraph;
}
