//////////////////////////////////////////////////////////////////////////
//
// FILE:   TextArea.C
//
// DATE:   January 3, 1996
//
// AUTHOR:
//
//    James Kuffner, Jr
//    Computer Science Department
//    Stanford University
//    U.S.A.
//    email: kuffner@stanford.edu
//    http://robotics.stanford.edu/~kuffner
//
// DISTRIBUTION NOTICE
//
// Copyright (c) 1996 James Kuffner, Jr. All rights reserved.
//
// Permission to use, copy, modify and distribute this software for any
// purpose is hereby granted without fee, provided that the above
// copyright notice and this permission notice appear in all copies of
// this software and that you do not sell the software.
//
// THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
//////////////////////////////////////////////////////////////////////////

#ifndef IGNORE_DEPEND
#include <math.h>
#include <stdio.h>
#include <Xm/Form.h>
#include <Xm/RowColumn.h>
#include <Xm/LabelG.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>
#include <Xm/Scale.h>
#endif

#include "TextArea.h"


#define MinHeight  100
#define MaxHeight  400


// string constants for sliders
static char* QuatNames[NUM_FIELDS] = {"q0","q1","q2","q3",
				      "Interpolation Weight"};
static char* AxisNames[NUM_FIELDS] = {"Angle","Axis X","Axis Y","Axis Z",
				      "Interpolation Weight"};
static char* EulerNames[NUM_FIELDS] ={"Roll","Pitch","Yaw","",
				      "Interpolation Weight"};


////////////////
//
// Construct a new textarea

TextArea::TextArea(Widget parent)
{
  _InitInterface(parent);
  for (int i = 0; i < TaNumCallbacks; _callbacks[i++] = NULL);
}

////////////////
//
// Initialize the X- and Motif-specific parts of the textarea.

void TextArea::_InitInterface(Widget parent)
{
  // record some X-related state
  this->_parent = parent;

  // create a form widget for the textarea elements
  _mainWindow = XtVaCreateWidget("MainWindow", xmFormWidgetClass, parent,
				XmNpaneMinimum, MinHeight,
				XmNpaneMaximum, MaxHeight,
				NULL);

  // create a row column widget for text
  Widget rowcol = XtVaCreateManagedWidget("rowcol", 
					  xmRowColumnWidgetClass,
					  _mainWindow,
					  XmNpacking, XmPACK_COLUMN,
					  XmNnumColumns, NUM_FIELDS,
					  XmNorientation, XmHORIZONTAL,
					  NULL);
  
  // add radio buttons
  Widget radiobox = XmCreateRadioBox(_mainWindow, "radio_box", NULL, 0);

  XtVaSetValues(rowcol,
		XmNleftAttachment, XmATTACH_FORM,
		XmNrightAttachment, XmATTACH_WIDGET,
		XmNrightWidget, radiobox,
		XmNtopAttachment, XmATTACH_FORM,
		XmNbottomAttachment, XmATTACH_FORM,
		NULL);
  
  XtVaSetValues(radiobox,
		XmNrightAttachment, XmATTACH_FORM,
		XmNbottomAttachment, XmATTACH_FORM,
		XmNtopAttachment, XmATTACH_FORM,
		NULL);

  _mode = QUAT;
  _radWidget[QUAT] = XtVaCreateManagedWidget("Quaternions              ",
					     xmToggleButtonGadgetClass,
					     radiobox, NULL);
  XtAddCallback(_radWidget[QUAT], XmNvalueChangedCallback, _RadioCB, this);
  
  _radWidget[AXIS] = XtVaCreateManagedWidget("Axis & Angle             ",
					     xmToggleButtonGadgetClass,
					     radiobox, NULL);
  XtAddCallback(_radWidget[AXIS], XmNvalueChangedCallback, _RadioCB, this);
  
  _radWidget[EULER] = XtVaCreateManagedWidget("XYZ Euler Angles        ",
					      xmToggleButtonGadgetClass,
					      radiobox, NULL);
  XtAddCallback(_radWidget[EULER], XmNvalueChangedCallback, _RadioCB, this);

  // create the slider widgets
  for (int i = 0; i < NUM_FIELDS; i++) {
    _edit[i] =XtVaCreateManagedWidget(QuatNames[i], xmScaleWidgetClass, rowcol,
				      XmNorientation, XmHORIZONTAL,
				      XmNshowValue, TRUE,
				      XmNwidth, 500,
				      NULL);
    
    XtAddCallback(_edit[i], XmNdragCallback, _TextCB, this);
    XtAddCallback(_edit[i], XmNvalueChangedCallback, _TextCB, this);
  }
  
  // realize the widgets
  XtManageChild(radiobox);  
  XtManageChild(rowcol);  
  XtManageChild(_mainWindow);  

  // make quaternions active
  XmToggleButtonSetState(_radWidget[QUAT],TRUE,FALSE);
}

////////////////
//
// Associate a function with a given callback

void TextArea::setCallback(TaCallbackType type, TaCallback callback,
			     TaPointer clientData)
{
  if (type >= 0 && type < TaNumCallbacks) {
    _callbacks[type] = callback;
    this->_clientData[type] = clientData;
  }
}

////////////////
// Execute a callback, if it has been set, and return its return
// value.

int TextArea::executeCallback(TaCallbackType type)
{
  if (type >= 0 && type < TaNumCallbacks)
    return _callbacks[type] ? (_callbacks[type])(this, this->_values,
						 _clientData[type]) : 1;
  else
    return 0;
}


////////////////
// SetSliderValues

void TextArea::SetSliderValues(float q[4])
{
  int value, i;

  switch (_mode) {
  case QUAT:
    for (i = 0; i < 4; i++) {
      value = rint(q[i] * 100);
      XtVaSetValues(_edit[i], XmNvalue, value, NULL);
    }
    break;

  case AXIS:
    for (i = 0; i < 3; i++) {
      value = rint(q[i] * 100);
      XtVaSetValues(_edit[i], XmNvalue, value, NULL);
    }
    value = rint(q[i] * 100);
    XtVaSetValues(_edit[i], XmNvalue, value, NULL);
    break;
    
  case EULER:
    for (i = 0; i < 3; i++) {
      value = rint(q[i] * 100);
      XtVaSetValues(_edit[i], XmNvalue, value, NULL);
    }
    break;

  default:
    fprintf(stderr, "unknown mode\n");
    break;
  }
}


////////////////
// GetSliderValues

void TextArea::GetSliderValues(float q[4])
{
  for (int i=0; i < 4; i++)
    q[i] = _values[i];
}


////////////////////////////////////////////////////////////////
// SetWeight
////////////////////////////////////////////////////////////////
void TextArea::SetWeight(float w)
{
  int value = rint(w * 100);
  XtVaSetValues(_edit[4], XmNvalue, value, NULL);
}

////////////////////////////////////////////////////////////////
// _InitMode
////////////////////////////////////////////////////////////////
void TextArea::InitMode()
{
  _mode = QUAT;
  _UpdateMode();
}


////////////////////////////////////////////////////////////////
// _RadioCB
////////////////////////////////////////////////////////////////

void TextArea::_RadioCB(Widget w, XtPointer clientData, XtPointer)
{
  TextArea *th = (TextArea *) clientData;
  
  // First turn all radio buttons to FALSE...
  XmToggleButtonSetState(th->_radWidget[QUAT],FALSE,FALSE);
  XmToggleButtonSetState(th->_radWidget[AXIS],FALSE,FALSE);
  XmToggleButtonSetState(th->_radWidget[EULER],FALSE,FALSE);
  
  // Set the appropriate radio button for the view
  XmToggleButtonSetState(w,TRUE,FALSE);
  
  // Set the mode
  for (int m = 0; m < NUM_RADIO_BUTTONS; m++)
    if (th->_radWidget[m] == w)
      th->_mode = (RotMode)m;
  th->_UpdateMode();
}


////////////////////////////////////////////////////////////////
// _TextCB
////////////////////////////////////////////////////////////////

void TextArea::_TextCB(Widget w, XtPointer clientData, XtPointer)
{
 TextArea *th = (TextArea *) clientData;
 int intval;

 //XmScaleCallbackStruct *cbs = (XmScaleCallbackStruct *)callData;

 // get the values of all of the sliders
 for (int i = 0; i < NUM_FIELDS; i++) {
   XmScaleGetValue(th->_edit[i], &intval);
   th->_values[i] = intval/100.0;
 }

 // if weight changed, reinterpolate
 if (th->_edit[P_WEIGHT] == w) {
   th->executeCallback(TaWeightChange);
   return; 
 }

 switch (th->_mode) {
 case QUAT:
   th->executeCallback(TaQuatChange);
   break;
   
 case AXIS:
   th->executeCallback(TaAxisChange);
   break;
   
 case EULER:
   for (i = 0; i < 3; i++) {
     XmScaleGetValue(th->_edit[i], &intval);
     th->_values[i] = intval/100.0;
   }
   th->executeCallback(TaEulerChange);
   break;
   
 default:
   fprintf(stderr, "unknown mode\n");
   break;
 }
}


////////////////////////////////////////////////////////////////
// _UpdateMode
////////////////////////////////////////////////////////////////
void TextArea::_UpdateMode()
{
  char **names;
  int max[NUM_FIELDS], min[NUM_FIELDS], val[NUM_FIELDS], dec[NUM_FIELDS];
  int i;

  // set the default values
  for (i = 0; i < NUM_FIELDS; i++) {
    min[i] = -100;
    max[i] = 100;
    val[i] = 0;
    dec[i] = 2;
  }
  min[P_WEIGHT] = 0;

  switch (_mode) {

  case QUAT:
    names = QuatNames;
    XtManageChild(_edit[NUM_FIELDS-2]);
    break;

  case AXIS:
    names = AxisNames;
    min[0] = -18000;
    max[0] = 18000;
    XtManageChild(_edit[NUM_FIELDS-2]);
    break;

  case EULER:
    names = EulerNames;
    for (i = 0; i < 3; i++) {
      min[i] = -18000;
      max[i] = 18000;
    }
    XtUnmanageChild(_edit[NUM_FIELDS-2]);
    break;

  default:
    fprintf(stderr, "Unknown mode!\n");
    break;
  }

  // set the widget arg values
  for (i = 0; i < NUM_FIELDS; i++) {
    XtVaSetValues(_edit[i],
		  XtVaTypedArg, XmNtitleString,
		  XmRString, names[i], strlen(names[i])+1,
		  XmNmaximum, max[i],
		  XmNminimum, min[i],
		  XmNvalue,   val[i],
		  XmNdecimalPoints, dec[i],
		  NULL);
  }

  executeCallback(TaUpdateValues);
}

